"""
A collection of classes to modify PMFs
"""

from .pmf import PMF, PMFCollection


# pylint: disable=R0201,C0302,R0913,R0904
class Modifier:
    """
    A base class used to represent a modfidier to a dice roll it includes
    the no-op stub methods for all dice operations.

    Attributes
    ----------
    priority : int
        Determines the priority order for the modifiers

    Methods
    -------
    modify_dice(col, thresh=None,  mod_thresh=None)
        Modifies the raw PMF

    modify_re_roll(col, thresh=None, mod_thresh=None)
        Modifies the dice through re-rolls

    modify_threshold(thresh)
        Modified the dice roll threshold in some way

    modify_save(save)
        Modify the save threshold in some way

    modify_ap(ap)
        Modify the armour penetration characteristic in some way

    modify_invuln(invuln)
        Modify the invulnerable save threshold in some way

    mod_extra_hit()
        Return the details for if a dice roll generates extra automatic hits
        that trigger on a modifiable threshold (eg: 5+)

    extra_hit()
        Return the details for if a dice roll generates extra automatic hits
        that trigger on an unmodifiable threshold (eg: 5 or 6)

    mod_extra_shot()
        Return the details for if a dice roll generates extra chances to hit
        that trigger on a modifiable threshold (eg: 5+)

    extra_shot()
        Return the details for if a dice roll generates extra chances to hit
        that trigger on an unmodifiable threshold (eg: 5 or 6)

    mod_extra_wound()
        Return the details for if a dice roll generates extra automatic wounds
        that trigger on a modifiable threshold (eg: 5+)

    extra_wound()
        Return the details for if a dice roll generates extra automatic wounds
        that trigger on an unmodifiable threshold (eg: 5 or 6)

    mod_mortal_wound()
        Return the details for if a dice roll generates mortal wounds that
        trigger on a modifiable threshold (eg: 5+)

    mortal_wound()
        Return the details for if a dice roll generates mortal wounds that
        trigger on an unmodifiable threshold (eg: 5 or 6)

    modify_drones()
        Return the details for if saviour protocols are active

    self_wound_thresh()
        Return the details for the to hit roll can generate wounds for the attacker
    """

    priority = 0
    # def __init__(self, keywords):
    #     self.keywords = keywords

    def modify_dice(self, col: PMFCollection, __: int, ___: int) -> PMFCollection:
        """Modifies the dice distribution. Sperate from the re-rolls becuse
        modifiers happen after re-rolls

        Parameters
        ----------
        col : PMFCollection
            The unmodified PMF collection that the modifier will be applied to
        thresh : int
            The unmodified dice success threshold, used to determine what values
            should be re-rolled. This is seperate from the modified threshold because
            "reroll failed" dice and "reroll all dice" interact with modifers
            differently
        mod_thresh : int, optional
            The modified dice success threshold, used to determine what values
            should be re-rolled (default is None)

        Returns
        -------
        PMFCollection
            The modified PMF Collection
        """
        return col

    def modify_re_roll(self, col: PMFCollection, _: int, __: int) -> PMFCollection:
        """Modifies the dice distribution via re-rolls

        Parameters
        ----------
        col : PMFCollection
            The unmodified PMF collection that the modifier will be applied to
        thresh : int
            The unmodified dice success threshold, used to determine what values
            should be re-rolled. This is seperate from the modified threshold because
            "reroll failed" dice and "reroll all dice" interact with modifers
            differently
        mod_thresh : int, optional
            The modified dice success threshold, used to determine what values
            should be re-rolled (default is None)

        Returns
        -------
        PMFCollection
            The modified PMF Collection
        """
        return col

    def modify_threshold(self, thresh: int) -> int:
        """Modifies the dice success threshold

        Parameters
        ----------
        thresh : int
            The unmodified dice success threshold

        Returns
        -------
        int
            Returns the provided dice success threshold in the base class
        """
        return thresh

    def modify_save(self, save: int) -> int:
        """Modifies the save success threshold

        Parameters
        ----------
        save : int
            The unmodified save dice success threshold

        Returns
        -------
        int
            Returns the provided save in the base class
        """
        return save

    def modify_ap(self, armour_penetration: int) -> int:
        """Modifies the armour penetration characteristic

        Parameters
        ----------
        armour_penetration : int
            The unmodified armour penetration characteristic

        Returns
        -------
        int
            Returns the provided armour penetration characteristic in the base class
        """
        return armour_penetration

    def modify_invuln(self, invuln: int) -> int:
        """Modifies the invunerable armour penetration characteristic

        Parameters
        ----------
        invuln : int
            The unmodified invunerable armour penetration characteristic

        Returns
        -------
        int
            Returns the provided invuln in the base class
        """
        return invuln

    def mod_extra_hit(self) -> PMFCollection:
        """Returns a PMF Collection of the modifiable extra hits generated

        Returns
        -------
        PMFCollection
            Returns None in the base class
        """
        return None

    def extra_hit(self) -> PMFCollection:
        """Returns a PMF Collection of the extra hits generated

        Returns
        -------
        PMFCollection
            Returns None in the base class
        """
        return None

    def mod_extra_shot(self) -> PMFCollection:
        """Returns a PMF Collection of the modifiable extra shots generated

        Returns
        -------
        PMFCollection
            Returns None in the base class
        """
        return None

    def extra_shot(self) -> PMFCollection:
        """Returns a PMF Collection of the modifiable extra shots generated

        Returns
        -------
        PMFCollection
            Returns None in the base class
        """
        return None

    def mod_extra_wound(self) -> PMFCollection:
        """Returns a PMF Collection of the modifiable extra shots generated

        Returns
        -------
        PMFCollection
            Returns None in the base class
        """
        return None

    def extra_wound(self) -> PMFCollection:
        """Returns a PMF Collection of the unmodifiable extra wounds generated

        Returns
        -------
        PMFCollection
            Returns None in the base class
        """
        return None

    def mod_mortal_wound(self) -> PMFCollection:
        """Returns a PMF Collection of the modifiable extra mortal wounds generated

        Returns
        -------
        PMFCollection
            Returns None in the base class
        """
        return None

    def mortal_wound(self) -> PMFCollection:
        """Returns a PMF Collection of the unmodifiable extra mortal wounds generated

        Returns
        -------
        PMFCollection
            Returns None in the base class
        """
        return None

    def modify_drones(self) -> tuple:
        """Returns a tuple of if saviour protocols is enabled and what the threshold for
        it is and the fnp value for the drones

        Returns
        -------
        tuple
            (False, 7, 7)
        """
        return False, 7, 7

    def self_wound_thresh(self) -> int:
        """Returns the to hit threshold for self wounds to generate below

        Returns
        -------
        int
            Base class always returns 0
        """
        return 0


class MinimumValue(Modifier):
    """
    A modifier to set a minimum value for a dice. For example some weapons that roll
    a D6 for the mount of damage state the values of one and two are considered threes.
    This adds the probability of rolling values less than the min_val to the probability
    for the min value

    Attributes
    min_val : int
        The minimum value that can be rolled on the dice

    Methods
    -------
    modify_dice(col: PMFCollection)
        Returns the PMFCollection where the PMFs have been modified to have a lower limit
    """

    def __init__(self, min_val: int, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.min_val = min_val

    def modify_dice(self, col: PMFCollection, *_) -> PMFCollection:
        return col.map(lambda x: x.min(self.min_val))


class ExplodingDice(Modifier):
    """
    The base class for the exploding dice modifiers.

    Attributes
    thresh : int
        The threshold value that extra dice are generated on

    value : int
        The number of extra dice generated

    Methods
    -------
    mod_extra_hit()
        Returns the PMFCollection for the exploding dice

    _pmf_collection()
        Returns the PMFCollection for the exploding dice
    """
    def __init__(self, thresh: int, value: int, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # clamp the threshold between 0 and 7
        self.thresh = max(0, min(7, thresh))
        self.value = value

    def _pmf_collection(self) -> PMFCollection:
        return PMFCollection(
            [PMF.static(0)] * self.thresh + [PMF.static(self.value)] * (7 - self.thresh)
        )


class ModExtraHit(ExplodingDice):
    """
    Exploding dice for generating extra extra automatic hits.

    Attributes
    thresh : int
        The threshold value that extra dice are generated on

    value : int
        The number of extra dice generated

    Methods
    -------
    mod_extra_hit()
        Returns the PMFCollection for the exploding dice
    """
    def mod_extra_hit(self) -> PMFCollection:
        return self._pmf_collection()


class ExtraHit(ExplodingDice):
    """
    Exploding dice for generating extra extra automatic hits on
    unmodified rolls.

    Attributes
    thresh : int
        The threshold value that extra dice are generated on

    value : int
        The number of extra dice generated

    Methods
    -------
    extra_hit()
        Returns the PMFCollection for the exploding dice
    """
    def extra_hit(self) -> PMFCollection:
        return self._pmf_collection()


class ModExtraAttack(ExplodingDice):
    """
    Exploding dice for generating extra extra attacks.

    Attributes
    thresh : int
        The threshold value that extra dice are generated on

    value : int
        The number of extra dice generated

    Methods
    -------
    mod_extra_shot()
        Returns the PMFCollection for the exploding dice
    """
    def mod_extra_shot(self) -> PMFCollection:
        return self._pmf_collection()


class ExtraShot(ExplodingDice):
    """
    Exploding dice for generating extra extra attacks on
    unmodified rolls.

    Attributes
    thresh : int
        The threshold value that extra dice are generated on

    value : int
        The number of extra dice generated

    Methods
    -------
    extra_shot()
        Returns the PMFCollection for the exploding dice
    """
    def extra_shot(self) -> PMFCollection:
        return self._pmf_collection()


class ModExtraWound(ExplodingDice):
    """
    Exploding dice for generating extra automatic wounds.

    Attributes
    thresh : int
        The threshold value that extra dice are generated on

    value : int
        The number of extra dice generated

    Methods
    -------
    mod_extra_wound()
        Returns the PMFCollection for the exploding dice
    """
    def mod_extra_wound(self) -> PMFCollection:
        return self._pmf_collection()


class ExtraWound(ExplodingDice):
    """
    Exploding dice for generating extra automatic wounds on
    unmodified rolls.

    Attributes
    thresh : int
        The threshold value that extra dice are generated on

    value : int
        The number of extra dice generated

    Methods
    -------
    extra_wound()
        Returns the PMFCollection for the exploding dice
    """
    def extra_wound(self) -> PMFCollection:
        return self._pmf_collection()


class GenerateMortalWound(ExplodingDice):
    """
    Exploding dice for generating mortal wounds on unmodified rolls.

    Attributes
    thresh : int
        The threshold value that extra dice are generated on

    value : int
        The number of extra dice generated

    Methods
    -------
    mortal_wound()
        Returns the PMFCollection for the exploding dice
    """
    def mortal_wound(self) -> PMFCollection:
        return self._pmf_collection()


class ModGenerateMortalWound(ExplodingDice):
    """
    Exploding dice for generating mortal wounds.

    Attributes
    thresh : int
        The threshold value that extra dice are generated on

    value : int
        The number of extra dice generated

    Methods
    -------
    mod_mortal_wound()
        Returns the PMFCollection for the exploding dice
    """
    def mod_mortal_wound(self) -> PMFCollection:
        return self._pmf_collection()


class Haywire(ExplodingDice):
    """
    Simulating the Haywire effects.

    Methods
    -------
    mod_mortal_wound()
        Returns the PMFCollection for the exploding dice
    """
    def mod_mortal_wound(self) -> PMFCollection:
        # This is a hardcoded implementation of the haywire mechanic of generating a mortal
        # wound on 4 and 5 and D3 mortal wounds on a 6+
        col = [
            PMF.static(0), # 0 for 0
            PMF.static(0), # 0 for 1
            PMF.static(0), # 0 for 2
            PMF.static(0), # 0 for 3
            PMF.static(1), # 1 for 4
            PMF.static(1), # 1 for 5
            PMF.dn(3), # d3 for 6
        ]
        return PMFCollection(col)


class ReRollOnes(Modifier):
    """
    Simulate rerolling the one value of the distributions

    Attributes
    ----------
    priority : int
        priority = 1

    Methods
    -------
    modify_re_roll()
        Returns the PMFCollection after re-rolling ones
    """
    priority = 1
    def modify_re_roll(self, col: PMFCollection, *_) -> PMFCollection:
        return col.map(lambda x: x.re_roll_value(1))


class ReRollFailed(Modifier):
    """
    Simulate rerolling the of failed rolls. Since this is only failed rolls before
    any modifiers it uses the worse value of the thresh ans mod thresh

    Attributes
    ----------
    priority : int
        priority = 99

    Methods
    -------
    modify_re_roll()
        Returns the PMFCollection after re-rolling failed rolls
    """
    priority = 99
    def modify_re_roll(self, col: PMFCollection, thresh: int, mod_thresh: int) -> PMFCollection:
        rr_thresh = min(thresh, mod_thresh)
        return col.map(lambda x: x.re_roll_less_than(rr_thresh))


class ReRollOneDice(Modifier):
    """
    Simulate rerolling the of all rolls of a single code in the PMF collection

    Methods
    -------
    modify_re_roll()
        Returns the PMFCollection after re-rolling all rolls
    """
    def modify_re_roll(self, col: PMFCollection, thresh: int, _) -> PMFCollection:
        pmfs = col.pmfs
        if not pmfs:
            return col
        pmfs[0] = pmfs[0].re_roll_less_than(thresh)
        return PMFCollection(pmfs)


class ReRollOneDiceVolume(Modifier):
    """
    Simulate rerolling one of the volume dice by rerolling vaues lower than the expected
    value of the dice.

    Methods
    -------
    modify_re_roll()
        Returns the PMFCollection after values lower than the epected value
    """
    def modify_re_roll(self, col: PMFCollection, *_) -> PMFCollection:
        pmfs = col.pmfs
        if not pmfs:
            return col
        pmfs[0] = pmfs[0].re_roll_less_than(pmfs[0].mean())
        return PMFCollection(pmfs)


class ReRollAll(Modifier):
    """
    Simulate rerolling all the dice. This allows you to include negative thrshold modifiers.

    Attributes
    ----------
    priority : int
        priority = 100

    Methods
    -------
    modify_re_roll()
        Returns the PMFCollection after rerolling all values
    """
    priority = 100
    def modify_re_roll(self, col: PMFCollection, _, mod_thresh: int) -> PMFCollection:
        return col.map(lambda x: x.re_roll_less_than(mod_thresh))


class ReRollLessThanExpectedValue(Modifier):
    """
    Simulate rerolling the volume dice by rerolling vaues lower than the expected
    value of the dice.

    Methods
    -------
    modify_re_roll()
        Returns the PMFCollection after values lower than the epected value
    """
    priority = 98
    def modify_re_roll(self, col: PMFCollection, *_) -> PMFCollection:
        return col.map(lambda x: x.re_roll_less_than(x.mean()))


class HighestOfTwo(Modifier):
    """
    Simulate rolling two dice and choosing the highest

    Methods
    -------
    modify_dice()
        Returns the PMFCollection of the highest of two dice
    """
    def modify_dice(self, col: PMFCollection, *_) -> PMFCollection:
        return col.map(lambda x: x.max_of_two())


class AddNTo(Modifier):
    """
    Base class for modifying dice or thresholds by a set amount
    """
    def __init__(self, value: int, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.value = value
        self.priority = self.value


class AddNToThreshold(AddNTo):
    """
    Add the value N to the threshold value

    Methods
    -------
    modify_threshold()
        Returns the modified threshold value
    """
    def modify_threshold(self, thresh: int) -> int:
        return thresh - self.value


class AddNToAP(AddNTo):
    """
    Add the value N to the armour penetraion value

    Methods
    -------
    modify_ap()
        Returns the modified armour penetration value
    """
    def modify_ap(self, armour_penetration: int) -> int:
        return armour_penetration - self.value


class AddNToSave(AddNTo):
    """
    Add the value N to the save value

    Methods
    -------
    modify_save()
        Returns the modified save value
    """
    def modify_save(self, save: int) -> int:
        return save - self.value


class AddNToInvuln(AddNTo):
    """
    Add the value N to the invunlerable save value

    Methods
    -------
    modify_invuln()
        Returns the modified save value
    """
    def modify_invuln(self, invuln: int) -> int:
        return invuln - self.value


class AddNToVolume(AddNTo):
    """
    Add the value N to the volume roll eg: d6 + 2

    Methods
    -------
    modify_dice()
        Returns the modified PMFCollection
    """
    def modify_dice(self, col: PMFCollection, *_) -> PMFCollection:
        return col.map(lambda x: x.roll(self.value))


class AddD6(AddNTo):
    """
    Adds a d6 to the volume roll

    Methods
    -------
    modify_dice()
        Returns the modified PMFCollection
    """
    def modify_dice(self, col: PMFCollection, *_) -> PMFCollection:
        return PMFCollection(col.pmfs+[PMF.dn(6)])


class AddD3(AddNTo):
    """
    Adds a d3 to the volume roll

    Methods
    -------
    modify_dice()
        Returns the modified PMFCollection
    """
    def modify_dice(self, col: PMFCollection, *_) -> PMFCollection:
        return PMFCollection(col.pmfs+[PMF.dn(3)])


class SetToN(Modifier):
    """
    Base class of modifiers that fix the PMF or threshold to a certain value
    """
    def __init__(self, value: int, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.value = value


class SetThresholdToN(SetToN):
    """
    Sets the threshold to N value

    Methods
    -------
    modify_threshold()
        Returns the N value
    """
    def modify_threshold(self, thresh: int) -> int:
        return self.value


class SetAPToN(SetToN):
    """
    Sets the armour pentration to N value

    Methods
    -------
    modify_ap()
        Returns the N value
    """
    def modify_ap(self, armour_penetration: int) -> int:
        return self.value


class SetSaveToN(SetToN):
    """
    Sets the save to N value

    Methods
    -------
    modify_save()
        Returns the N value
    """
    def modify_save(self, save: int) -> int:
        return self.value


class SetInvulnToN(SetToN):
    """
    Sets the invunerable save to N value

    Methods
    -------
    modify_invuln()
        Returns the N value
    """
    def modify_invuln(self, invuln: int) -> int:
        return self.value


class IgnoreAP(Modifier):
    """
    If the value is lower than N then set the ap to zero

    Methods
    -------
    modify_ap()
        Returns the N value
    """
    def __init__(self, value: int, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.armour_penetration = value

    def modify_ap(self, armour_penetration: int) -> int:
        return 0 if armour_penetration <= self.armour_penetration else armour_penetration


class IgnoreInvuln(Modifier):
    """
    Set the invunlerable to 7

    Methods
    -------
    modify_invuln()
        Returns 7
    """
    def modify_invuln(self, _: int) -> int:
        return 7


class HalfDamage(Modifier):
    """
    Modify the volume rolls to half the value rounding up

    Methods
    -------
    modify_invuln()
        Returns The modified PMFCollection
    """
    def modify_dice(self, col: PMFCollection, *_) -> PMFCollection:
        return col.map(lambda x: x.div_min_one(2))


class ShieldDrone(Modifier):
    """
    Mark that saviour protocols are active with a threshold of 2+ and a fnp of 5+++

    Methods
    -------
    modify_drones()
        Returns drone info
    """
    def modify_drones(self) -> tuple:
        return True, 2, 5


class NormalDrone(Modifier):
    """
    Mark that saviour protocols are active with a threshold of 2+ and a fnp of 7+++

    Methods
    -------
    modify_drones()
        Returns drone info
    """
    def modify_drones(self) -> tuple:
        return True, 2, 7


class Overheat(Modifier):
    """
    Mark that the attack inflicts damage on the attacker

    Methods
    -------
    modify_drones()
        Return a 2+ threshold to not inflict wounds
    """
    def self_wound_thresh(self) -> int:
        return 2


class ModifierCollection:
    """
    Used to keep track of any modifiers to the attack
    """
    def __init__(self, shot_mods=None, hit_mods=None, wound_mods=None, pen_mods=None,
                 fnp_mods=None, damage_mods=None):

        self.shot_mods = self._sort_priority(shot_mods or [])
        self.hit_mods = self._sort_priority(hit_mods or [])
        self.wound_mods = self._sort_priority(wound_mods or [])
        self.pen_mods = self._sort_priority(pen_mods or [])
        self.fnp_mods = self._sort_priority(fnp_mods or [])
        self.damage_mods = self._sort_priority(damage_mods or [])

    def __add__(self, other):
        # Useful when you want to add two ModifierCollection together
        if not isinstance(other, ModifierCollection):
            raise TypeError(f'{other} is not ModifierCollection')

        return ModifierCollection(
            shot_mods=self.shot_mods+other.shot_mods,
            hit_mods=self.hit_mods+other.hit_mods,
            wound_mods=self.wound_mods+other.wound_mods,
            pen_mods=self.pen_mods+other.pen_mods,
            fnp_mods=self.fnp_mods+other.fnp_mods,
            damage_mods=self.damage_mods+other.damage_mods,
        )

    def _sort_priority(self, mods: list) -> list:
        return sorted(mods, key=lambda x: x.priority, reverse=True)

    def _mod_dice(self, col: PMFCollection, mods: list, thresh=None,
                  mod_thresh=None) -> PMFCollection:
        """Modifies the dice distribution

        Parameters
        ----------
        col : PMFCollection
            The unmodified PMF collection that the modifier will be applied to
        mods : list
            The list of mods to be applied
        thresh : list
            The unmodified dice success threshold, used to determine what values
            should be re-rolled. This is seperate from the modified threshold because
            "reroll failed" dice and "reroll all dice" interact with modifers
            differently
        mod_thresh : int, optional
            The modified dice success threshold, used to determine what values
            should be re-rolled (default is None)

        Returns
        -------
        PMFCollection
            The modified PMF Collection
        """
        for mod in mods:
            col = mod.modify_re_roll(col, thresh, mod_thresh)
        for mod in mods:
            col = mod.modify_dice(col, thresh, mod_thresh)
        return col

    def modify_shot_dice(self, col: PMFCollection) -> PMFCollection:
        """
        Modify the PMF of shot volume the dice. Ususally for re-rolls.
        """
        return self._mod_dice(col, self.shot_mods)

    def modify_hit_thresh(self, thresh: int) -> int:
        """
        Modify the hit threshold. Important to note the -1 to hit modifiers actually
        are a +1 to the threshold. Similarly +1 to hits are -1 to the threshold.
        """
        if thresh == 1:
            # Handle the case where the weapon is auto hit. No to hit modifiers map
            return thresh
        for mod in self.hit_mods:
            thresh = mod.modify_threshold(thresh)
        return max(thresh, 2) #1's always fail

    def modify_hit_dice(self, col: PMFCollection, thresh: int, mod_thresh: int) -> PMFCollection:
        """
        Modify the PMF of hit dice. Ususally for re-rolls.
        """
        return self._mod_dice(col, self.hit_mods, thresh, mod_thresh)

    def modify_wound_thresh(self, thresh: int) -> int:
        """
        Modify the wound threshold. Important to note the -1 to wound modifiers actually
        are a +1 to the threshold. Similarly +1 are -1 to the threshold.
        """
        for mod in self.wound_mods:
            thresh = mod.modify_threshold(thresh)
        return max(2, thresh) # 1's always fail

    def modify_wound_dice(self, dists, thresh, mod_thresh):
        """
        Modify the PMF of hit dice. Ususally for re-rolls.
        """
        return self._mod_dice(dists, self.wound_mods, thresh, mod_thresh)

    def modify_pen_thresh(self, save, armour_penetration, invuln):
        """
        Modify the pen threshold by modifying the save, ap, and invuln
        """
        for mod in self.pen_mods:
            save = mod.modify_save(save)
        for mod in self.pen_mods:
            armour_penetration = mod.modify_ap(armour_penetration)
        for mod in self.pen_mods:
            invuln = mod.modify_invuln(invuln)
        return min(max(save + armour_penetration, 2), max(invuln, 2)) # 1's alwasys fail

    def modify_pen_dice(self, dists, thresh, mod_thresh):
        """
        Modify the PMF of the pen dice. Ususally for re-rolls.
        """
        return self._mod_dice(dists, self.pen_mods, thresh, mod_thresh)

    def modify_drone(self):
        """
        Return if the attack should be modified by saviour protocols
        """
        enabled = False
        thresh = 7
        fnp = 7
        for mod in self.pen_mods:
            mod_enabled, mod_threshold, mod_fnp = mod.modify_drones()

            enabled = enabled or mod_enabled
            thresh = thresh if thresh < mod_threshold else mod_threshold
            fnp = fnp if fnp < mod_fnp else mod_fnp

        return enabled, thresh, fnp

    def self_wound_thresh(self) -> int:
        """
        Return threshold for self wound
        """
        thresh = 0
        for mod in self.hit_mods:
            mod_thresh = mod.self_wound_thresh()
            thresh = thresh if thresh > mod_thresh else mod_thresh
        return thresh


    def modify_fnp_thresh(self, thresh):
        """
        Modify the fnp threshold. I think some death guard units can do this?
        """
        for mod in self.fnp_mods:
            thresh = mod.modify_save(thresh)
        return max(thresh, 2) # 1's alwasys fail

    def modify_fnp_dice(self, dists, thresh, mod_thresh):
        """
        Modify the PMF of the FNP dice. Ususally for re-rolls.
        """
        return self._mod_dice(dists, self.fnp_mods, thresh, mod_thresh)

    def modify_damage_dice(self, dists):
        """
        Modify the damage dice
        """
        return self._mod_dice(dists, self.damage_mods)

    def sum_generators(self, mod_list, attr_name):
        """
        Sum the generators
        """
        cols = [getattr(mod, attr_name)() for mod in mod_list]
        cols = [x for x in cols if x]
        return PMFCollection.add_many(cols)

    def get_mod_extra_hit(self):
        """
        Generate extra hits on a modfiable value
        """
        return self.sum_generators(self.hit_mods, 'mod_extra_hit')

    def get_extra_hit(self) -> PMFCollection:
        """
        Generate extra hits on a static value
        """
        return self.sum_generators(self.hit_mods, 'extra_hit')

    def get_mod_extra_shot(self) -> PMFCollection:
        """
        Generate extra shots on a modfiable value
        """
        return self.sum_generators(self.hit_mods, 'mod_extra_shot')

    def get_extra_shot(self) -> PMFCollection:
        """
        Generate extra shots on a static value
        """
        return self.sum_generators(self.hit_mods, 'extra_shot')

    def get_mod_extra_wound(self) -> PMFCollection:
        """
        Generate extra wounds on a modfiable value
        """
        return self.sum_generators(self.wound_mods, 'mod_extra_wound')

    def get_extra_wound(self) -> PMFCollection:
        """
        Generate extra wounds on a static value
        """
        return self.sum_generators(self.wound_mods, 'extra_wound')

    def hit_mod_mortal_wounds(self) -> PMFCollection:
        """
        Generate mortal wounds on a modfiable value
        """
        return self.sum_generators(self.hit_mods, 'mod_mortal_wound')

    def wound_mod_mortal_wounds(self) -> PMFCollection:
        """
        Generate mortal wounds on a modfiable value
        """
        return self.sum_generators(self.wound_mods, 'mod_mortal_wound')

    def hit_mortal_wounds(self) -> PMFCollection:
        """
        Generate mortal wounds on a static value
        """
        return self.sum_generators(self.hit_mods, 'mortal_wound')

    def wound_mortal_wounds(self) -> PMFCollection:
        """
        Generate mortal wounds on a static value
        """
        return self.sum_generators(self.wound_mods, 'mortal_wound')



# -*- coding: utf-8 -*-
"""Warhammer Stats

This module included the statistics components from the https://www.warhammer-stats-engine.com/
website.

"""

from .attack import Attack
from .target import Target
from .weapon import Weapon
from .pmf import PMF, PMFCollection



"""
Classes related to simulating an attack in Warhammer 40000 8th edition
"""

from __future__ import annotations

from .pmf import PMF, PMFCollection
from .weapon import Weapon
from .target import Target
from .modifiers import ModifierCollection



# pylint: disable=R0201,C0302,R0913,R0902,R0903,R0904,R0913

class AttackResults:
    """Holds the results of an attack sequence.

    Args:
        shot_volume_results (PMF): The results from determining the shot volume
        to_hit_results (PMF): The results from the to-hit phase
        to_wound_results (PMF): The results from the to-wound phase
        saviour_protocol_results (PMF): The results from applying saviour protocols
        armour_save_results (PMF): The results from the saving throws phase
        damage_results (PMF): The results from appling wounds phase

    """
    def __init__(self, shot_volume_results: ShotVolumePhaseResults,
                 to_hit_results: ToHitPhaseResults,
                 to_wound_results: ToWoundPhaseResults,
                 saviour_protocol_results: SaviourProtocolPhaseResults,
                 armour_save_results: ArmourSavePhaseResults,
                 damage_results: DamagePhaseResults):

        self.shot_volume_results = shot_volume_results
        self.to_hit_results = to_hit_results
        self.to_wound_results = to_wound_results
        self.saviour_protocol_results = saviour_protocol_results
        self.armour_save_results = armour_save_results
        self.damage_results = damage_results

    @property
    def damage_dist(self) -> PMF:
        """
        The final damage distribution
        """
        return self.damage_results.damage_dist

    @property
    def mortal_wound_dist(self) -> PMF:
        """
        The combined damage distribution from the hit and wound phase
        """
        return PMF.convolve_many([
            self.to_hit_results.mortal_wound_dist,
            self.to_wound_results.mortal_wound_dist,
        ])

    @property
    def total_wounds_dist(self) -> PMF:
        """
        The combined damage distribution from regular and mortal wounds
        """
        return PMF.convolve_many([
            self.mortal_wound_dist,
            self.damage_dist,
        ])

class ShotVolumePhaseResults:
    """Holds the results of determining the number of attacks.

    Args:
        shot_dist (PMF): The distribution of shots to be made
    """
    def __init__(self, shots_dist: PMF):
        self.shots_dist = shots_dist


class ToHitPhaseResults:
    """Holds the results from the to hit phase of the attack

    Args:
        hit_dist (PMF): The distribution of successful hits
        exploding_dice_dists (PMF): The distribution of successful hits generated from
            exploding dice
        mortal_wound_dist (PMF): The distribution of mortal wounds generated
        self_inflicted_dist (PMF): The distribution of wounds inflicted on the attacker
    """
    def __init__(self, hit_dist: PMF, exploding_dice_dist: PMF, mortal_wound_dist: PMF,
                 self_inflicted_dist: PMF):
        self.hit_dist = hit_dist
        self.exploding_dice_dist = exploding_dice_dist
        self.mortal_wound_dist = mortal_wound_dist
        self.self_inflicted_dist = self_inflicted_dist

    @property
    def combined_hit_dists(self) -> PMF:
        """
        The combined damage distribution from the hit and exploding wounds
        """
        return PMF.convolve_many([self.hit_dist, self.exploding_dice_dist])

    @classmethod
    def empty(cls) -> ToHitPhaseResults:
        """
        Return an empty set of results
        """
        return ToHitPhaseResults(
            hit_dist=PMF.static(0),
            exploding_dice_dist=PMF.static(0),
            mortal_wound_dist=PMF.static(0),
            self_inflicted_dist=PMF.static(0),
        )


class ToWoundPhaseResults:
    """Holds the results from the to wound phase of the attack

    Args:
        wounds_dist (PMF): The distribution of successful wounds
        mortal_wound_dist (PMF): The distribution of mortal wounds generated
    """
    def __init__(self, wound_dist: PMF, exploding_dice_dist: PMF, mortal_wound_dist: PMF):
        self.wound_dist = wound_dist
        self.exploding_dice_dist = exploding_dice_dist
        self.mortal_wound_dist = mortal_wound_dist


class SaviourProtocolPhaseResults:
    """Holds the results from saviour protocols

    Args:
        wounds_dist (PMF): The distribution of successful wounds
        mortal_wound_dist (PMF): The distribution of mortal wounds generated
    """
    def __init__(self, failed_save_dist: PMF, drone_wound_dist: PMF):
        self.failed_save_dist = failed_save_dist
        self.drone_wound_dist = drone_wound_dist


class ArmourSavePhaseResults:
    """Holds the results from the armour save phase

    Args:
        wounds_dist (PMF): The distribution of successful wounds
        mortal_wound_dist (PMF): The distribution of mortal wounds generated
    """
    def __init__(self, failed_save_dist: PMF):
        self.failed_save_dist = failed_save_dist


class DamagePhaseResults:
    """Holds the results from the armour save phase

    Args:
        wounds_dist (PMF): The distribution of successful wounds
        mortal_wound_dist (PMF): The distribution of mortal wounds generated
    """
    def __init__(self, damage_dist: PMF):
        self.damage_dist = damage_dist


class Attack:
    """Generates the probability distribution for damage dealt from an attack sequence

    Note:
        This can be pretty computaionally expensive to run. Don't go crazy with the number
        attacks and expect it to run quickly.

    Args:
        weapon (Weapon): The weapon being used to make the attack
        target (Target): The target of the the attack

    Attributes:
        msg (str): Human readable string describing the exception.
        code (int): Exception error code.
    """
    def __init__(self, weapon: Weapon, target: Target):
        self.weapon = weapon
        self.target = target

    def _shot_volume_phase(self):
        return ShotVolumePhase(self.weapon, self.target, self.modifiers)

    def _to_hit_phase(self):
        return ToHitPhase(self.weapon, self.target, self.modifiers)

    def _to_wound_phase(self):
        return ToWoundPhase(self.weapon, self.target, self.modifiers)

    def _saviour_protocol_phase(self):
        return SaviourProtocolPhase(self.weapon, self.target, self.modifiers)

    def _armour_save_phase(self):
        return ArmourSavePhase(self.weapon, self.target, self.modifiers)

    def _damage_phase_phase(self):
        return DamagePhase(self.weapon, self.target, self.modifiers)

    @property
    def modifiers(self) -> ModifierCollection:
        """Return a combined list of modifers for both the weapon and the target"""
        return self.weapon.modifiers + self.target.modifiers

    def run(self) -> AttackResults:
        """
        Generate the resulting PMF
        """


        # Generate the PMF for the number of shots
        shot_volume_results = self._shot_volume_phase().calc_dist()

        # Generate the PMFs for the number of hits, mortal wounds, and self inflicted
        # mortal wounds in the hit phase
        to_hit_results = self._to_hit_phase().calc_dist(
            dist=shot_volume_results.shots_dist,
            can_recurse=True,
        )

        # Generate the PMFs for the number of wounds, mortal wounds, and self inflicted
        # mortal wounds in the wound phase
        to_wound_results = self._to_wound_phase().calc_dist(dist=to_hit_results.combined_hit_dists)

        # Generate the PMFs for the number of wounds that penetrate saviour protocols and
        # how many drones die
        saviour_protocol_results = self._saviour_protocol_phase().calc_dist(
            to_wound_results.wound_dist,
        )

        # Generate the PMFs for the number of failed saves in the armor save phase
        armour_save_results = self._armour_save_phase().calc_dist(
            saviour_protocol_results.failed_save_dist,
        )

        # Generate the PMFs for the danage dealt in the damage calculation phase.
        damage_results = self._damage_phase_phase().calc_dist(
            armour_save_results.failed_save_dist,
        )

        return AttackResults(
            shot_volume_results=shot_volume_results,
            to_hit_results=to_hit_results,
            to_wound_results=to_wound_results,
            saviour_protocol_results=saviour_protocol_results,
            armour_save_results=armour_save_results,
            damage_results=damage_results,
        )


class AttackPhase:
    """The base class for attack segments

    Note:
        This is just a base class that contains a number of common methods

    Args:
        attack (Attack): A reference to the attack being made
    """
    def __init__(self, weapon: Weapon, target: Target, modifiers: ModifierCollection):
        self.weapon = weapon
        self.target = target
        self.modifiers = modifiers
        self._thresh_mod = None

    @property
    def thresh_mod(self) -> int:
        """
        A cached copy of the threshold modifier
        """
        # cache the threshold modifier
        if self._thresh_mod is None:
            self._thresh_mod = self._get_thresh_mod()
        return self._thresh_mod

    def _get_thresh_mod(self) -> int:
        return 0

    def _calc_exp_dist(self, dice_dists: PMF) -> PMF:
        # Calculate the exploding dice distributions
        dist_col = PMFCollection.add_many([
            self._mod_extra_dist().thresh_mod(self.thresh_mod),
            self._extra_dist(),
        ])

        if dist_col:
            return dist_col.mul_col(dice_dists).convolve()
        return PMF.static(0)

    def _mod_extra_dist(self) -> PMFCollection:
        # Stub
        return PMFCollection.empty()

    def _extra_dist(self) -> PMFCollection:
        # Stub
        return PMFCollection.empty()

    def _calc_mortal_dist(self, dice_dists) -> PMF:
        # Calculate how many mortal wounds are generated
        dist_col = PMFCollection.add_many([
            self._mod_mortal_dist().thresh_mod(self.thresh_mod),
            self._mortal_dist(),
        ])
        if dist_col:
            return dist_col.mul_col(dice_dists).convolve()
        return PMF.static(0)

    def _mod_mortal_dist(self) -> PMFCollection:
        # Stub
        return PMFCollection.empty()

    def _mortal_dist(self) -> PMFCollection:
        # Stub
        return PMFCollection.empty()


class ShotVolumePhase(AttackPhase):
    """
    Generate the PMF for the number of shots
    """
    def calc_dist(self, *_) -> ShotVolumePhaseResults:
        """Calculate the probability distiribution for the number of shots made in the attack
        for weapons with a fixed value this is a static number but some can be the convolution
        of more than one PMF. eg: 'roll two d6'
        """
        # Apply the shot volume modifiers
        shots_dists = self.modifiers.modify_shot_dice(self.weapon.shots)

        # Convolve the resulting PMFs
        return ShotVolumePhaseResults(
            shots_dist=shots_dists.convolve()
        )


class ToHitPhase(AttackPhase):
    """
    Simulate the hits phase of the attack
    """
    def calc_dist(self, dist: PMF, can_recurse: bool, *_) -> ToHitPhaseResults: # pylint: disable=too-many-locals
        """Calculate the probability distiribution of the number of sucessful hits in
        in the attack phase. This takes into account modifiers and other effects as
        extra hits and shots.
        """
        hit_dists = []
        exploding_dice_dists = []
        mortal_wound_dists = []
        self_inflicted_dists = []

        # The bare threshold to hit is the weapons to hit value
        thresh = self.weapon.bs

        # Check for any modifers to the to hit roll
        mod_thresh = self.modifiers.modify_hit_thresh(self.weapon.bs)

        # Check for the threshhold for self inflicted wounds
        thresh_self_wounds = self.modifiers.self_wound_thresh()

        # Generate the resulting hit distributon for each value of the shots PMF then
        # multipy by the probability of that value and sum them
        for dice_count, event_prob in enumerate(dist.values):
            # If the probability is zero then no-op
            if event_prob == 0:
                continue

            # Create a dice distribution for 'dice_count' d6 dice
            dice_dists = PMFCollection.mdn(dice_count, 6)

            # Apply modifiers to the dice distribution
            dice_dists = self.modifiers.modify_hit_dice(dice_dists, thresh, mod_thresh)

            # If self inflicted wounds are possible calculate them
            if thresh_self_wounds:
                self_thresh = max(thresh_self_wounds+mod_thresh-thresh, 0)
                self_inflicted_dist = dice_dists.convert_binomial_less_than(self_thresh).convolve()
            else:
                self_inflicted_dist = PMF.static(0)


            # Convert the dice dist into a binomial distribution with the modified to hit value.
            # We are bsically flatteing all the individual d6 down into d2 True/False for if it hit.
            hit_dist = dice_dists.convert_binomial(mod_thresh).convolve()

            # Calculate the distirbution for exploding automatic hits. Since extra shots cannot
            # generate more hits return a zero prob if we are recusing
            exp_dist = self._calc_exp_dist(dice_dists) if can_recurse else PMF.static(0)

            # Calculate the ditributons for exploding extra chances to hit. This requires us
            # to calcuate. The whole to-hit phase again for the sub hits
            explosive_shot_result = self._calc_exp_shot_dist(dice_dists, can_recurse)

            # Calculate the mortal wounds gnerated
            mrt_dist = self._calc_mortal_dist(dice_dists)

            # Multiply the resulting hit distributions by the probability of this number of dice
            # and append them to the totals
            hit_dists.append(hit_dist * event_prob)
            combined_exploding_dice = PMF.convolve_many([exp_dist, explosive_shot_result.hit_dist])
            combined_mortal_wounds = PMF.convolve_many(
                [mrt_dist, explosive_shot_result.mortal_wound_dist],
            )
            combined_self_inflicted = PMF.convolve_many(
                [self_inflicted_dist, explosive_shot_result.self_inflicted_dist],
            )

            # Add these to the list
            exploding_dice_dists.append(combined_exploding_dice * event_prob)
            mortal_wound_dists.append(combined_mortal_wounds * event_prob)
            self_inflicted_dists.append(combined_self_inflicted * event_prob)

        # Flattening the dists means we are summing the probabilites of each value across all the
        # dists. This is how we find the total probability distribution of the phase.

        return ToHitPhaseResults(
            hit_dist=PMF.flatten(hit_dists),
            exploding_dice_dist=PMF.flatten(exploding_dice_dists),
            mortal_wound_dist=PMF.flatten(mortal_wound_dists),
            self_inflicted_dist=PMF.static(0) if not self_inflicted_dists else \
                PMF.flatten(self_inflicted_dists)
        )

    def _get_thresh_mod(self):
        # This is a bit of a hack to get the delta between the threshold and the modified threshold
        return self.modifiers.modify_hit_thresh(6) - 6

    def _calc_exp_shot_dist(self, dice_dists: PMF, can_recurse: bool) -> ToHitPhaseResults:
        # This handles generating extra chances to hit and then calculating the distribution of
        # resulting successful hits

        # If we are alread recusing then return zeros
        if not can_recurse:
            return ToHitPhaseResults.empty()

        # Since to hit modifers can modify the number of extra chances to hit generate that now
        dist_col = PMFCollection.add_many([
            self._get_mod_extra_shot().thresh_mod(self.thresh_mod),
            self._get_extra_shot(),
        ])

        if dist_col:
            return self.calc_dist(dist_col.mul_col(dice_dists).convolve(), can_recurse=False)
        return ToHitPhaseResults.empty()

    def _mod_extra_dist(self):
        return self.modifiers.get_mod_extra_hit()

    def _extra_dist(self):
        return self.modifiers.get_extra_hit()

    def _get_mod_extra_shot(self):
        return self.modifiers.get_mod_extra_shot()

    def _get_extra_shot(self):
        return self.modifiers.get_extra_shot()

    def _mod_mortal_dist(self):
        return self.modifiers.hit_mod_mortal_wounds()

    def _mortal_dist(self):
        return self.modifiers.hit_mortal_wounds()


class ToWoundPhase(AttackPhase):
    """
    Generate the PMF for the wounds
    """
    def calc_dist(self, dist: PMF, *_) -> ToWoundPhaseResults:
        """Calculate the probability distiribution of the number of sucessful wounds in
        in the wound phase. This takes into account modifiers and other effects as
        extra automatic wounds.
        """
        wnd_dists = []
        exploding_dice_dists = []
        mortal_wound_dist = []

        # Calculate the wound threshold from the strength of the weapon and the toughness
        # of the target
        thresh = self._calc_wound_thresh(
            self.weapon.strength,
            self.target.toughness
        )

        # Apply any modifers to the wound threshold
        mod_thresh = self.modifiers.modify_wound_thresh(thresh)

        # Generate the resulting hit distributon for each value of the shots PMF then
        # multipy by the probability of that value and sum them
        for dice_count, event_prob in enumerate(dist.values):
            # If the probability is zero then no-op
            if event_prob == 0:
                continue

            # Create a dice distribution for 'dice_count' d6 dice
            dice_dists = PMFCollection.mdn(dice_count, 6)

            # Apply modifiers to the dice distribution
            dice_dists = self.modifiers.modify_wound_dice(dice_dists, thresh, mod_thresh)

            # Convert the dice dist into a binomial distribution with the modified to wound value.
            # We are bsically flatteing all the individual d6 down into d2 True/False for if it
            # wounded.
            wnd_dist = dice_dists.convert_binomial(mod_thresh).convolve()

            # Calculate exploding auto-wounds
            exp_dist = self._calc_exp_dist(dice_dists)

            # Generate mortal wounds
            mrt_dist = self._calc_mortal_dist(dice_dists)

            wnd_dists.append(wnd_dist * event_prob)
            exploding_dice_dists.append(exp_dist * event_prob)
            mortal_wound_dist.append(mrt_dist * event_prob)

        # Flatten the distros down
        return ToWoundPhaseResults(
            wound_dist=PMF.flatten(wnd_dists),
            exploding_dice_dist=PMF.flatten(exploding_dice_dists),
            mortal_wound_dist=PMF.flatten(mortal_wound_dist),
        )

    def _calc_wound_thresh(self, strength: int, toughness: int) -> int:
        # Generate the wound threshold
        if strength <= toughness/2.0: # pylint: disable=no-else-return
            # If the toughness is greater than twice the strength wound on a 6+
            return 6
        elif strength >= toughness*2:
            # If the stength is greater than twice the toughness wound on a 2+
            return 2
        elif toughness > strength:
            # Else if toughness is greater than strength wound on a 5_
            return 5
        elif toughness == strength:
            # Else if they are equal then wound on a 4+
            return 4
        else:
            # If the strength is greater than toughness wound on a 3+
            return 3

    def _get_thresh_mod(self) -> int:
        return self.modifiers.modify_wound_thresh(6) - 6

    def _mod_extra_dist(self) -> PMFCollection:
        return self.modifiers.get_mod_extra_wound()

    def _extra_dist(self) -> PMFCollection:
        return self.modifiers.get_extra_wound()

    def _mod_mortal_dist(self) -> PMFCollection:
        return self.modifiers.wound_mod_mortal_wounds()

    def _mortal_dist(self) -> PMFCollection:
        return self.modifiers.wound_mortal_wounds()


class SaviourProtocolPhase(AttackPhase):
    """
    Generate the PMF for saviour protocols
    """
    def calc_dist(self, dist: PMF) -> SaviourProtocolPhaseResults:
        """Calculate the probability distiribution of the number of failed saves from
        saviour protocol and the amount of damage dealt to drones. This takes into account \
        modifiers such as the shield drone FNP and other effects.
        """
        # Check if saviour protcols are active and if so what the threshold is and the FNP value
        drone_enabled, drone_thresh, drone_fnp = self.modifiers.modify_drone()

        # If the probability is zero then no-op
        if not drone_enabled:
            return SaviourProtocolPhaseResults(
                failed_save_dist=dist,
                drone_wound_dist=PMF.static(0),
            )

        # Calculate how many wounds pass though the savious protocols and how many wounds the
        # drones take
        pens = []
        saves = []
        for dice_count, event_prob in enumerate(dist.values):
            # If the probability is zero then no-op
            if event_prob == 0:
                continue

            # Create a dice distribution for 'dice_count' d6 dice
            dice_dists = PMFCollection.mdn(dice_count, 6)

            # Convert binomial for values less than the threshold for attacks the penetrated
            pen_dist = dice_dists.convert_binomial_less_than(drone_thresh).convolve()

            # Convert binomial for values greater than the threshold for wounds passed off to the
            # drones
            save_dist = dice_dists.convert_binomial(drone_thresh).convolve()

            pens.append(pen_dist * event_prob)
            saves.append(save_dist * event_prob)

        # Flatten the distros down
        return SaviourProtocolPhaseResults(
            failed_save_dist=PMF.flatten(pens),
            drone_wound_dist=self._calc_fnp_dist(PMF.flatten(saves), drone_fnp),
        )

    def _calc_fnp_dist(self, dist, drone_fnp) -> PMF:
        # Check how many of the wounds are removed by the drone FNP
        dists = []
        for dice, event_prob in enumerate(dist.values):
            if event_prob == 0:
                continue
            dice_dists = PMFCollection.mdn(dice, 6)
            binom_dists = dice_dists.convert_binomial_less_than(drone_fnp).convolve()
            dists.append(binom_dists * event_prob)
        return PMF.flatten(dists)


class ArmourSavePhase(AttackPhase):
    """
    Generate the PMF for the wounds that penetrate the
    """
    def calc_dist(self, dist: PMF) -> ArmourSavePhaseResults:
        """Calculate the probability distiribution of the number of failed armour
        saves. This takes into account modifiers and invunerable save
        """
        # Calculate the final save threshold from the save value, invunerable save, and
        # any save modifiers
        mod_thresh = self.modifiers.modify_pen_thresh(
            self.target.save,
            self.weapon.ap,
            self.target.invuln
        )

        dists = []
        for dice_count, event_prob in enumerate(dist.values):
            # If the probability is zero then no-op
            if event_prob == 0:
                continue

            # Create a dice distribution for 'dice_count' d6 dice
            dice_dists = PMFCollection.mdn(dice_count, 6)
            dice_dists = self.modifiers.modify_pen_dice(dice_dists, mod_thresh, mod_thresh)

            # Convert binomial for values less than the threshold for attacks the penetrated
            pen_dists = dice_dists.convert_binomial_less_than(mod_thresh).convolve()
            dists.append(pen_dists * event_prob)

        # Flatten the distros down
        return ArmourSavePhaseResults(
            failed_save_dist=PMF.flatten(dists)
        )


class DamagePhase(AttackPhase):
    """
    Generate the PMF for the damage dealt to the target
    """
    def calc_dist(self, dist: PMF) -> DamagePhaseResults:
        """Calculate the probability distiribution of the number of wounds dealt from a
        failed daving throw. This accounts for feel no pain, target wounds characteristic
        and other damage modifiers.

        Note: This does not take into account the inneficientcy from partially wounding
        units
        """
        # Get the distribution of each individual wound dealt
        individual_dist = self._calc_individual_damage_dist()

        damge_dists = []
        for dice, event_prob in enumerate(dist.values):
            # If the probability is zero then no-op
            if event_prob == 0:
                continue
            # Multiply the individual damage distributions by the number dice
            damge_dists.append(PMF.convolve_many([individual_dist] * dice) * event_prob)

        return DamagePhaseResults(
            damage_dist=PMF.flatten(damge_dists)
        )

    def _calc_individual_damage_dist(self) -> PMF:
        # Get the basic damage distribution
        dice_dists = self.weapon.damage

        # Apply modifers to the damage distribution
        mod_dists = self.modifiers.modify_damage_dice(dice_dists)

        # Calculate the damge distribution after the feel no pain has been applied
        fnp_dist = self._calc_fnp_dist(mod_dists.convolve())

        # Clip the damage dist to the wounds characteristic of the target
        return fnp_dist.ceiling(self.target.wounds)

    def _calc_fnp_dist(self, dist: PMF) -> PMF:
        dists = []
        mod_thresh = self.modifiers.modify_fnp_thresh(self.target.fnp)
        for dice, event_prob in enumerate(dist.values):
            if event_prob == 0:
                continue
            dice_dists = self.modifiers.modify_fnp_dice(
                PMFCollection.mdn(dice, 6),
                self.target.fnp,
                mod_thresh,
            )
            binom_dists = dice_dists.convert_binomial_less_than(mod_thresh).convolve()
            dists.append(binom_dists * event_prob)
        return PMF.flatten(dists)



"""
Classes related to modeling the target of an attack
"""

from .modifiers import ModifierCollection

# pylint: disable=too-many-arguments,too-few-public-methods

class Target:
    """
    Holds params of a target
    """
    def __init__(self, toughness: int, save: int, invuln: int, fnp: int, wounds: int,
                 modifiers: ModifierCollection = None):
        self.toughness = toughness
        self.save = save
        self.invuln = invuln
        self.fnp = fnp
        self.wounds = wounds
        self.modifiers = modifiers or ModifierCollection()



import setuptools

from pipenv.project import Project
from pipenv.utils import convert_deps_to_pip

pfile = Project(chdir=False).parsed_pipfile
requirements = convert_deps_to_pip(pfile['packages'], r=False)
test_requirements = convert_deps_to_pip(pfile['dev-packages'], r=False)

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='warhammer_stats',
    version="0.0.7",
    author="Adam Kabbeke",
    author_email='akabbeke@gmail.com',
    description="Warhammer 40k stats tools",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/akabbeke/warhammer_stats",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=requirements,
    test_requirements=test_requirements,
    python_requires='>=3.6',
)


"""
Classes related to modeling a probability mass function and collections of them
https://en.wikipedia.org/wiki/Probability_mass_function
"""

from __future__ import annotations
import math

from typing import List, Union, Callable
from numpy import fft, zeros

# pylint: disable=too-many-public-methods

class PMF:
    """
    Discrete Probability Distribution - Used to keep track of the probability of random discrete
    events
    """
    def __init__(self, values: List[float]):
        self.values = list(values)

    def __str__(self) -> str:
        return str(self.values)

    def __len__(self) -> int:
        return len(self.values)

    def __mul__(self, other: Union[int, float]) -> PMF:
        return PMF([other*x for x in self.values])

    def __rmul__(self, other: Union[int, float]) -> PMF:
        return self.__mul__(other)

    def ceiling(self, value: Union[int, float]) -> PMF:
        """
        Sum the probability of all values >= the ceiling value
        """
        new_values = []
        for i, prob in enumerate(self.values):
            if i <= value:
                new_values.append(prob)
            else:
                new_values[value] += prob
        return PMF(new_values)

    def trim_tail(self, thresh: Union[int, float]) -> PMF:
        """
        Trim the long tail off where p < threshold
        """
        in_body = False
        trimmed_dist = []
        for element in self.values[::-1]:
            if in_body or element >= thresh:
                in_body = True
                trimmed_dist.append(element)
        return PMF(trimmed_dist[::-1])

    def cumulative(self) -> PMF:
        """
        Probability of at least each value
        """
        cumu_dist = []
        for i in range(len(self.values)):
            cumu_dist.append(sum(self.values[i:]))
        return PMF(cumu_dist)

    def re_roll_value(self, value: Union[int, float]) -> PMF:
        """
        Re-roll a specific value (eg re-roll 1's)
        """
        rr_mask = [0.0 if x == value else 1.0 for x in range(len(self))]
        rr_dist = [self.values[value] * x for x in self.values]
        new_dist = []
        for i, prob in enumerate(self.values):
            new_dist.append(rr_mask[i]*prob + rr_dist[i])
        return PMF(new_dist)

    def re_roll_less_than(self, value: Union[int, float]) -> PMF:
        """
        Re-roll all values below a specific value
        """
        rr_mask = [0.0 if i < value else x for i, x in enumerate(self.values)]
        for i, _ in enumerate(self.values):
            if i < value:
                for j, _ in enumerate(self.values):
                    rr_mask[j] += self.values[i] * self.values[j]
        return PMF(rr_mask)

    def convert_binomial(self, thresh: Union[int, float]) -> PMF:
        """
        Convert the pmf into a binomial PMF by flatteing all values above or equal to the
        threshold into 1 and all below into 0
        """
        return PMF([sum(self.values[:thresh]), sum(self.values[thresh:])])

    def convert_binomial_less_than(self, thresh: Union[int, float]) -> PMF:
        """
        Convert the pmf into a binomial PMF by flatteing all values above or equal to the
        threshold into 0 and all below into 1
        """
        return PMF([sum(self.values[thresh:]), sum(self.values[:thresh])])

    def get(self, value: int) -> float:
        """
        Fetch the probability for the value.
        """
        if value < 0:
            return 0.0
        if value > len(self.values):
            return 0.0
        return self.values[value]

    def expand_to(self, length: int) -> PMF:
        """
        Pad values with zeros to reach the desired length
        """
        return PMF(self.values + [0.0] * max(length - len(self), 0))

    def add_value(self, value: int) -> PMF:
        """
        Add an integer value to the PMF by shifting the values right
        """
        return PMF([0.0] * value + self.values)

    def max_of_two(self) -> PMF:
        """
        Produce the PMF from rolling two of this PMF and choosing the higher
        """
        return PMF.max_of_two_pmf(self, self)

    def roll(self, roll_value: int) -> PMF:
        """
        Roll the values of the PMF left or right. If rolling right pad the left with zeroes.
        If rolling left flatten values into the zero value.
        """
        if roll_value == 0: # pylint: disable=no-else-return
            return self
        elif roll_value > 0:
            return PMF([0.0] * roll_value + self.values)
        return PMF([sum(self.values[:(-1*roll_value)+1])] + self.values[(-1*roll_value)+1:])

    def div_min_one(self, divisor: Union[int, float]) -> PMF:
        """
        Divide the index values of the PMF by the divisor with a minimum of one.
        I added this to accomidate Abadons half damage ability.
        """
        new_dist = [0.0] * len(self.values)
        for i, value in enumerate(self.values):
            new_dist[math.ceil(i/divisor)] += value
        return PMF(new_dist)

    def min(self, min_val: int) -> PMF:
        """
        Sets the minimum value of the PMF by adding the sum of all probabilites less than
        the min_val to the min val.
        """
        return PMF([0.0] * min_val + [sum(self.values[:min_val+1])] + self.values[min_val+1:])

    def mean(self) -> PMF:
        """
        Return the expected value of the PMF
        """
        return sum([x*p for x, p in enumerate(self.values)])

    def std(self) -> PMF:
        """
        Return the standard deviation of the PMF
        """
        mean = self.mean()
        exp_mean = sum([(x**2)*p for x, p in enumerate(self.values)])
        return (exp_mean - mean**2)**(0.5)

    @classmethod
    def dn(cls, dice_sides: int) -> PMF: # pylint: disable=invalid-name
        """
        Return the PMD for a dice with dice_sides number of sides
        """
        return PMF([0.0] + [1/dice_sides] * dice_sides)

    @classmethod
    def static(cls, static_value: int) -> PMF:
        """
        Return the PMD for exactly the static_value
        """
        return PMF([0.0] * static_value + [1.0])

    @classmethod
    def convolve_many(cls, dists: List[PMF]) -> PMF:
        """
        Convolve a list of 1d float arrays together, using FFTs.
        The arrays need not have the same length, but each array should
        have length at least 1.

        This works becuse the convolution in the frequency space is a single
        element-wise multiplication. FFT gets us into the frequency space faster
        than the full FT
        """

        result_length = 1 + sum((len(dist) - 1) for dist in dists)

        # Copy each array into a 2d array of the appropriate shape.
        rows = zeros((len(dists), result_length))
        for i, dist in enumerate(dists):
            rows[i, :len(dist)] = dist.values

        # Transform, take the product, and do the inverse transform
        # to get the convolution.
        fft_of_rows = fft.fft(rows)
        fft_of_convolution = fft_of_rows.prod(axis=0)
        convolution = fft.ifft(fft_of_convolution)

        # Assuming real inputs, the imaginary part of the output can
        # be ignored.
        return PMF(convolution.real)

    @classmethod
    def flatten(cls, dists: List[PMF]) -> PMF:
        """
        Sum a set of distributions to produce a new distribution.
        """
        flat_dist = [0.0] * (max([len(dist) for dist in dists]))
        for dist in dists:
            for i, value in enumerate(dist.values):
                flat_dist[i] += value
        return PMF(flat_dist)

    @classmethod
    def match_sizes(cls, dists: List[PMF]) -> PMF:
        """
        Map a set of PMFs to the same size
        """
        largest = max([len(x) for x in dists])
        return [x.expand_to(largest) for x in dists]

    @classmethod
    def max_of_two_pmf(cls, dist1: PMF, dist2: PMF) -> PMF:
        """
        Compute the PMF for the max of two PMF
        """
        dist1, dist2 = cls.match_sizes([dist1, dist2])
        new_dist = []
        for value in range(len(dist1)):
            prob = dist1.get(value) * dist2.get(value)
            prob += dist1.values[value] * sum(dist2.values[:value])
            prob += dist2.values[value] * sum(dist1.values[:value])
            new_dist.append(prob)
        return PMF(new_dist)

class PMFCollection:
    """
    Discrete Probability Distribution - Used to keep track of collections of PMFs
    """
    def __init__(self, pmfs: List[PMF] = None):
        self.pmfs = pmfs or []

    def __bool__(self) -> bool:
        return len(self) > 0

    def __len__(self) -> int:
        return len(self.pmfs)

    def __str__(self) -> str:
        return str([x.values for x in self.pmfs])

    def get(self, index: int, defualt: PMF = None) -> PMF:
        """
        Fetch the PMF at the index
        """
        try:
            return self.pmfs[index]
        except IndexError:
            return defualt

    def thresh_mod(self, thresh_mod: int) -> PMFCollection:
        """
        Modify the collection based on a dice modifer
        """
        if thresh_mod == 0 or self.pmfs == []: # pylint: disable=no-else-return
            return self
        elif thresh_mod > 0:
            return PMFCollection(self.pmfs[:1] * thresh_mod + self.pmfs)
        return PMFCollection(
            (self.pmfs + self.pmfs[-1:] * (-1 * thresh_mod))[-1*len(self.pmfs):]
        )

    def mul_col(self, other_collection: PMFCollection) -> PMFCollection:
        """
        Multipy this PMFCollection byt another PMFCollection
        """
        return PMFCollection([self.mul_pmf(pmf) for pmf in other_collection.pmfs])

    def mul_pmf(self, pmf: PMF) -> PMF:
        """
        Multiply the collection with a pmf and return a convolution of the results
        """
        new_pmfs = []
        for i, value in enumerate(pmf.values):
            new_pmfs.append(self.get(i, PMF.static(0)) * value)
        return PMF.flatten(new_pmfs)

    def convolve(self) -> PMF:
        """
        Convolve the PMFs into a single PMF
        """
        return PMF.convolve_many(self.pmfs)

    def convert_binomial_less_than(self, thresh: int) -> PMFCollection:
        """
        Convert the PMFs into a Binomial distribution for values less than the thresh
        """
        return self.map(lambda x: x.convert_binomial_less_than(thresh))

    def convert_binomial(self, thresh) -> PMFCollection:
        """
        Convert the PMFs into a Binomial distribution for values greater than or equal to the thresh
        """
        return self.map(lambda x: x.convert_binomial(thresh))

    def map(self, func: Callable[[PMF], PMF]) -> PMFCollection:
        """
        Map a function over the PMFs in the PMFCollection
        """
        return PMFCollection([func(x) for x in self.pmfs])

    @classmethod
    def add_many(cls, collection_list: List[PMFCollection]) -> PMFCollection:
        """
        Adds many PMFCollection
        """
        new_pmfs = []
        for i in range(max([len(x) for x in collection_list] or [0])):
            new_pmfs.append(
                PMF.convolve_many([x.get(i, PMF.static(0)) for x in collection_list if x])
            )
        return PMFCollection(new_pmfs)

    @classmethod
    def empty(cls) -> PMFCollection:
        """
        Return an empty PMFCollection
        """
        return PMFCollection([])

    @classmethod
    def mdn(cls, number_of_dice: int, number_of_sides: int) -> PMFCollection:
        """
        Return a PMFCollection of m dn PMFs
        """
        return PMFCollection([PMF.dn(number_of_sides)] * number_of_dice)

    @classmethod
    def static(cls, static_value) -> PMFCollection:
        """
        Return a PMFCollection of a single static value
        """
        return PMFCollection([PMF.static(static_value)])



"""
Classes for representing an attacking weapon
"""

from .pmf import PMFCollection
from .modifiers import ModifierCollection

# pylint: disable=too-many-arguments,too-few-public-methods

class Weapon:
    """
    Holds the params of a weapon
    """
    def __init__(self, bs: int, shots: PMFCollection, strength: int, ap: int,
                 damage: PMFCollection, modifiers: ModifierCollection = None):
        self.bs = bs # pylint: disable=invalid-name
        self.shots = shots
        self.strength = strength
        self.ap = ap # pylint: disable=invalid-name
        self.damage = damage
        self.modifiers = modifiers or ModifierCollection()
